(define-module (ordo connection)
  #:use-module (oop goops)
  #:use-module (ice-9 exceptions)
  #:use-module (ice-9 match)
  #:use-module (logging logger)
  #:use-module (srfi srfi-1)  ; list operations
  #:use-module (srfi srfi-71) ; extended let
  #:use-module (ordo connection base)
  #:use-module (ordo connection local)
  #:use-module (ordo connection ssh)
  #:use-module (ordo connection sudo)
  #:use-module (ordo util flatten)
  #:use-module (ordo util shell-quote)
  #:use-module (ordo util keyword-args)
  #:export (connection?
            local-connection
            ssh-connection
            call-with-connection
            run)
  #:re-export (conn:setup conn:teardown))

(define (connection? c)
  (is-a? c <connection>))

(define (local-connection)
  (make <local-connection>))

(define* (ssh-connection user host #:key (password #f) (identity #f) (authenticate-server? #t))
  (make <ssh-connection> #:user user #:host host #:password password
        #:identity identity #:authenticate-server? authenticate-server?))

(define* (call-with-connection c sudo? sudo-user sudo-password proc)
  (let ((c (if sudo?
               (make <sudo-connection> #:connection c #:become-user sudo-user #:become-password sudo-password)
               c)))
    (dynamic-wind
      (lambda () (conn:setup c))
      (lambda () (proc c))
      (lambda () (conn:teardown c)))))

(define (build-command prog args pwd env redirect-err?)
  (let ((xs (remove unspecified?
                    (flatten (list "env"
                                   (when pwd (list "--chdir" (string-shell-quote pwd)))
                                   (when env (map (match-lambda ((k . v) (string-append k "=" (string-shell-quote v)))) env))
                                   prog
                                   (map string-shell-quote args)
                                   (when redirect-err? "2>&1"))))))
    (string-join xs " ")))

(define (run conn prog . args)
  (let* ((args (flatten args))
         (args kwargs (break keyword? args))
         (args (remove unspecified? args))
         (pwd (keyword-arg kwargs #:pwd))
         (env (keyword-arg kwargs #:env))
         (return (keyword-arg kwargs #:return identity))
         (check? (keyword-arg kwargs #:check?))
         (command (build-command prog args pwd env #t)))
    (log-msg 'INFO "Running command: " command)
    (let ((out rc (conn:run conn command)))
      (log-msg 'INFO "Command exit code: " rc)
      (if check?
          (if (zero? rc)
              (return out)
              (raise-exception (make-exception
                                (make-external-error)
                                (make-exception-with-message (format #f "Non-zero exit (~a) from ~a" rc prog)))))
          (values (return out) rc)))))
